
(* Modify only typecheck and interpret *)

open Ast

exception TypeError
exception RunTimeError
exception Unimplemented (* shows you what to change *)

(*
(* commented out, but an approach equivalent to the one below using
   explicit mutation
*)
(*********** environment implementation where env = var->exp **********)

(* Note: types do not matter at run-time, so we just use Ignore to fill 
   in for a type we do not care about *) 

let mt_env x = raise RunTimeError

let extend_env env x e = fun y -> if x=y then e else env y

(* call this from the Rec case of interpret *)
let extend_env_rec env f x e = 
  let rec new_env y =
    if y=f
    then Closure(Fun(x,Ignore,e), new_env)
    else env y
  in new_env

let lookup_env env y = env y
*)

(******* environment implementation where env = (var * (exp ref)) list *******)

(* Note: types do not matter at run-time, so we just use Ignore to fill 
   in for a type we do not care about *) 

let mt_env = []

let extend_env env x e = (x,ref e)::env

(* call this from the Rec case of interpret *)
let extend_env_rec env f x e =
  let r       = ref True in
  let new_env = (f,r)::env in
  let _       = r := Closure(Fun(x,Ignore,e),new_env) in
  new_env

let rec lookup_env env y =
  match env with
    [] -> raise RunTimeError
  | (x,r)::tl -> if x=y then !r else lookup_env tl y

(* type-checker -- we have done a couple cases for you *)
(* Note: t1=t2 will do what you want to compare types structurally, 
   e.g., Arrow(Int,Bool)=Arrow(Int,Bool)
   But t1==t2 is roughly pointer-equality -- do not accidentally use it or
   you will be very sorry *)

let mt_ctxt = []
let extend_ctxt ctxt x t = (x,t)::ctxt
let rec lookup_ctxt ctxt y =
  match ctxt with
    [] -> raise TypeError
  | (x,t)::tl -> if x=y then t else lookup_ctxt tl y

let rec typecheck ctxt e =
  match e with
    Var(x) -> lookup_ctxt ctxt x
  | True  -> Bool
  | False -> Bool
  | If(e1,e2,e3) ->
      let t1 = typecheck ctxt e1 in
      let t2 = typecheck ctxt e2 in
      let t3 = typecheck ctxt e3 in
      if (t1=Bool) && (t2=t3)
      then t2
      else raise TypeError
  | MakePair(e1,e2) ->
      let t1 = typecheck ctxt e1 in
      let t2 = typecheck ctxt e2 in
      Pair(t1,t2)
  | First(e1) ->
      (match typecheck ctxt e1 with
	Pair(t1,t2) -> t1
      |	_ -> raise TypeError)
  | Second(e1) ->
      (match typecheck ctxt e1 with
	Pair(t1,t2) -> t2
      |	_ -> raise TypeError)
  | Fun(x,t,e) ->
      let t1 = typecheck (extend_ctxt ctxt x t) e in
      Arrow(t,t1)
  | Apply(e1,e2) ->
      let t1 = typecheck ctxt e1 in
      let t2 = typecheck ctxt e2 in
      (match t1 with
	Arrow(t3,t4) -> if t3=t2 then t4 else raise TypeError
      |	_ -> raise TypeError)
  | Rec(f,t,x,e) ->
      (match t with
	Arrow(t1,t2) -> 
	  let c1 = extend_ctxt ctxt f t  in
	  let c2 = extend_ctxt c1 x t1 in
	  let t3 = typecheck c2 e in
	  if t3=t2
	  then t
	  else raise TypeError
      |	_ -> raise TypeError)
  | Closure(_,_) -> raise TypeError

(* evaluator 
   remember we are large-step -- return a value
   *)

type version = PartC | PartD | PartE | PartF

let rec interpret v env e =
  let interp = interpret v in
  match e with
    Var(x) -> lookup_env env x
  | True -> e
  | False -> e
  | If(e1,e2,e3) ->
      (match interp env e1 with
	True -> interp env e2
      |	False -> interp env e3
      |	_ -> raise RunTimeError)
  | MakePair(e1,e2) -> 
      (match v with
	PartC -> MakePair(interp env e1, interp env e2)
      |	PartD -> 
	  let r = ref None in
	  let _ = Thread.create (fun () -> r := Some(interp env e2)) () in
	  let v1 = interp env e1 in
	  let rec spin () =
	    (match !r with
	      Some v2 -> MakePair(v1,v2)
	    | _ -> Thread.yield(); spin())
	  in spin()
      |	PartE ->
	  let r = ref None in
	  let m = Mutex.create() in
	  let c = Condition.create() in
	  let _ = Thread.create 
	      (fun () -> 
	         Mutex.lock m; 
		 r := Some(interp env e2); 
		 Condition.signal c;
                 Mutex.unlock m) 
               () in
	  let v1 = interp env e1 in
	  let _ = Mutex.lock m in
	  let rec spin () =
	    (match !r with
	      Some v2 -> MakePair(v1,v2)
	    | _ -> Condition.wait c m; spin()) in
	  spin ()
      |	PartF -> 
	  let r = ref None in
	  let tid = Thread.create (fun () -> r := Some(interp env e2)) () in
	  let v1 = interp env e1 in
	  let _ = Thread.join tid in
	  (match !r with
	    Some v2 -> MakePair(v1,v2)
	  | _ -> failwith "impossible -- I joined"))
  | First(e1) ->
      (match interp env e1 with
	MakePair(x,y) -> x
      |	_ -> raise RunTimeError)
  | Second(e1) ->
      (match interp env e1 with
	MakePair(x,y) -> y
      |	_ -> raise RunTimeError)
  | Closure(_,_) -> e
  | Fun(_,_,_) -> Closure(e,env)
  | Apply(e1,e2) ->
      let v1 = interp env e1 in
      let v2 = interp env e2 in
      (match v1 with
	Closure(Fun(x,_,e3),envc) -> interp (extend_env envc x v2) e3
      |	_ -> raise RunTimeError)
  | Rec(f,t,x,e) ->
      Closure(Fun(x,Ignore,e),extend_env_rec env f x e)

(*Do not change anything below here*)

(* expect one command-line argument, a file to parse *)
(* you do not need to understand this interaction with the system *)
let parse_file () =
  let argv = Sys.argv in
  let _ = 
    if Array.length argv != 2
    then (prerr_string ("usage: " ^ argv.(0) ^ " [file-to-parse]\n");
	  exit 1) in
  let ch = open_in argv.(1) in
  Parse.program Lex.lexer (Lexing.from_channel ch)

let _ =
  let prog = parse_file() in
  (try
    (ignore(typecheck mt_ctxt prog);
(*     print_string (exp2string prog); *) (* useful for debugging *)
     print_string ("\ninterpreted answer = ");
     print_string (exp2string (interpret PartC mt_env prog));
     print_string ("\ninterpreted answer2 = ");
     print_string (exp2string (interpret PartD mt_env prog));
     print_string ("\ninterpreted answer3 = ");
     print_string (exp2string (interpret PartE mt_env prog));
     print_string ("\ninterpreted answer4 = ");
     print_string (exp2string (interpret PartF mt_env prog));
     print_newline ())
  with TypeError -> print_string "did not typecheck\n")
